@@ -42,7 +42,6 @@ use Lintian::Check qw(check_spelling spelling_tag_emitter);
 use Lintian::Data;
 use Lintian::Relation qw(:constants);
 use Lintian::Tags qw(tag);
-use Lintian::Output qw(debug_msg);
 use Lintian::Util qw(fail slurp_entire_file strip);
 
 my $ARCH_REGEX = Lintian::Data->new('binaries/arch-regex', qr/\s*\~\~/o,
@@ -99,6 +98,7 @@ sub _split_hash {
 
 our $HARDENING= Lintian::Data->new('binaries/hardening-tags', qr/\s*\|\|\s*/o,
     \&_split_hash);
+our $HARDENED_FUNCTIONS = Lintian::Data->new('binaries/hardened-functions');
 our $LFS_SYMBOLS = Lintian::Data->new('binaries/lfs-symbols');
 
 our $ARCH_32_REGEX;
@@ -107,8 +107,7 @@ sub run {
     my ($pkg, $type, $info, $proc, $group) = @_;
 
     my ($madir, %directories, $built_with_golang, %SONAME);
-    my $gnu_triplet_re;
-    my $ruby_triplet_re;
+    my ($arch_hardening, $gnu_triplet_re, $ruby_triplet_re);
     my $needs_libc = '';
     my $needs_libcxx = '';
     my $needs_libc_file;
@@ -124,9 +123,12 @@ sub run {
     my $multiarch = $info->field('multi-arch', 'no');
     my $srcpkg = $proc->pkg_src;
 
+    $arch_hardening = $HARDENING->value($arch)
+      if $arch ne 'all';
+
     foreach my $file (sort keys %{$info->objdump_info}) {
         my $objdump = $info->objdump_info->{$file};
-        my $has_lfs;
+        my ($has_lfs, %unharded_functions, @hardened_functions);
         my $is_profiled = 0;
         # $file can be an object inside a static lib.  These do
         # not appear in the output of our file_info collection.
@@ -152,6 +154,20 @@ sub run {
         foreach my $symbol (@{$objdump->{SYMBOLS}}) {
             my ($foo, $sec, $sym) = @{$symbol};
 
+            if ($foo eq 'UND') {
+                my $name = $sym;
+                my $hardened;
+                $hardened = 1 if $name =~ s/^__(\S+)_chk$/$1/;
+                if ($HARDENED_FUNCTIONS->known($name)) {
+                    if ($hardened) {
+                        push(@hardened_functions, $name);
+                    } else {
+                        $unharded_functions{$name} = 1;
+                    }
+                }
+
+            }
+
             unless (defined $has_lfs) {
                 if ($foo eq 'UND' and $LFS_SYMBOLS->known($sym)) {
                     # Using a 32bit only interface call, some parts of the
@@ -178,6 +194,13 @@ sub run {
             tag 'binary-compiled-with-profiling-enabled', $file
               if $is_profiled;
         }
+        if (    %unharded_functions
+            and not @hardened_functions
+            and not $built_with_golang
+            and $arch_hardening->{'hardening-no-fortify-functions'}) {
+            tag 'hardening-no-fortify-functions', $file;
+        }
+
         tag 'apparently-corrupted-elf-binary', $file
           if $objdump->{'ERRORS'};
         tag 'binary-file-built-without-LFS-support', $file
@@ -230,7 +253,7 @@ sub run {
     $base_pkg =~ s/c2a?\b//o;
     $base_pkg =~ s/\dg$//o;
     $base_pkg =~ s/gf$//o;
-    $base_pkg =~ s/v5$//o; # GCC-5 / libstdc++6 C11 ABI breakage
+    $base_pkg =~ s/v[5-6]$//o; # GCC-5 / libstdc++6 C11 ABI breakage
     $base_pkg =~ s/-udeb$//o;
     $base_pkg =~ s/^lib64/lib/o;
 
@@ -250,10 +273,11 @@ sub run {
     tag 'package-name-doesnt-match-sonames', "@sonames"
       if @sonames && !$match_found;
 
-    my $src = $group->get_source_processable();
+    my $src = $group->get_source_processable;
     if (defined($src)) {
         $built_with_golang
-          = $src->info->relation('build-depends')->implies('golang-go');
+          = $src->info->relation('build-depends')
+          ->implies('golang-go | golang-any');
     }
 
     # process all files in package
@@ -360,11 +384,11 @@ sub run {
         };
         my $tag_emitter
           = spelling_tag_emitter('spelling-error-in-binary', $file);
-        check_spelling($strings, $exceptions, $tag_emitter);
+        check_spelling($strings, $exceptions, $tag_emitter, 0);
 
         # stripped?
         if ($fileinfo =~ m,\bnot stripped\b,o) {
-            # Is it an object file (which generally can not be
+            # Is it an object file (which generally cannot be
             # stripped), a kernel module, debugging symbols, or
             # perhaps a debugging package?
             unless ($fname =~ m,\.k?o$,
@@ -392,7 +416,7 @@ sub run {
         }
 
         # rpath is disallowed, except in private directories
-        if (exists $objdump->{RPATH}) {
+        if (exists($objdump->{RPATH}) or exists($objdump->{RUNPATH})) {
             if (not %directories) {
                 for my $file ($info->sorted_index) {
                     my $name;
@@ -402,11 +426,10 @@ sub run {
                     $directories{"/$name"}++;
                 }
             }
+            my @rpaths
+              = (keys(%{$objdump->{RPATH}}),keys(%{$objdump->{RUNPATH}}),);
 
-            foreach my $rpath (
-                map {File::Spec->canonpath($_)}
-                keys %{$objdump->{RPATH}}
-              ) {
+            foreach my $rpath (map {File::Spec->canonpath($_)}@rpaths) {
                 next
                   if $rpath
                   =~ m,^/usr/lib/(?:$madir/)?(?:games/)?(?:\Q$pkg\E|\Q$srcpkg\E)(?:/|\z),;
@@ -553,23 +576,20 @@ sub run {
                 }
             }
 
-            # Check for missing hardening characteristics. This currently
-            # handles the following checks:
-            # no-relro no-fortify-functions no-stackprotector no-bindnow no-pie
-            if (exists($info->hardening_info->{$fname})) {
-                my $flags = $HARDENING->value($arch);
-                if ($flags) {
-                    foreach my $t (@{$info->hardening_info->{$fname}}) {
-                        my $tag = "hardening-$t";
-                        # Binaries built by the Go compiler do not support all
-                        # hardening measures.
-                        next
-                          if ($t eq 'no-relro'
-                            ||$t eq 'no-fortify-functions')
-                          &&$built_with_golang;
-                        tag $tag, $file if $flags->{$tag};
-                    }
-                }
+            if (    $arch_hardening->{'hardening-no-relro'}
+                and not $built_with_golang
+                and not $objdump->{'PH'}{'RELRO'}) {
+                tag 'hardening-no-relro', $file;
+            }
+
+            if ($arch_hardening->{'hardening-no-bindnow'}
+                and not exists($objdump->{'FLAGS_1'}{'NOW'})) {
+                tag 'hardening-no-bindnow', $file;
+            }
+
+            if (    $arch_hardening->{'hardening-no-pie'}
+                and $objdump->{'ELF-TYPE'} eq 'EXEC') {
+                tag 'hardening-no-pie', $file;
             }
         }
     }
@@ -615,7 +635,7 @@ sub run {
     if ($has_perl_lib) {
         # It is a virtual package, so no version is allowed and
         # alternatives probably does not make sense here either.
-        my $re = qr/^perlapi-[\d.]+(?:\s*\[[^\]]+\])?$/;
+        my $re = qr/^perlapi-[-\w.]+(?:\s*\[[^\]]+\])?$/;
         unless ($depends->matches($re, VISIT_OR_CLAUSE_FULL)) {
             tag 'missing-dependency-on-perlapi';
         }
