@@ -33,7 +33,7 @@ use Lintian::Check qw($known_shells_regex);
 use Lintian::Data;
 use Lintian::Relation;
 use Lintian::Tags qw(tag);
-use Lintian::Util qw(fail strip);
+use Lintian::Util qw(do_fork fail strip);
 
 # This is a map of all known interpreters.  The key is the interpreter
 # name (the binary invoked on the #! line).  The value is an anonymous
@@ -54,7 +54,7 @@ my $INTERPRETERS = Lintian::Data->new('scripts/interpreters', qr/\s*=\>\s*/o,
 # dashes, digits, and periods.
 #
 # The list contains the following values:
-#  [<path>, <dependency-relation>, <regex>, <dependency-templatel>, <version-list>]
+#  [<path>, <dependency-relation>, <regex>, <dependency-template>, <version-list>]
 #
 # Their meaning is documented in Lintian's scripts/versioned-interpreters
 # file, though they are ordered differently and there are a few differences
@@ -298,6 +298,10 @@ sub run {
             && !$is_absolute
             && $in_examples);
 
+        # Skip upstream source code shipped in /usr/share/cargo/registry/
+        next
+          if $filename =~ m,^usr/share/cargo/registry/,;
+
         if ($interpreter eq '') {
             script_tag('script-without-interpreter', $filename);
             next;
@@ -371,22 +375,33 @@ sub run {
             }
         } elsif ($interpreter =~ m,/usr/local/,) {
             script_tag('interpreter-in-usr-local', $filename,"#!$interpreter");
-        } elsif ($executable{'.' . $interpreter}) {
-         # Package installs the interpreter itself, so it's probably ok.  Don't
-         # emit any tag for this.
         } elsif ($interpreter eq '/bin/env') {
             script_tag('script-uses-bin-env', $filename);
+        } elsif ($base =~ /^php/) {
+            script_tag('php-script-with-unusual-interpreter',
+                $filename, "$interpreter");
+
+            # This allows us to still perform the dependencies checks
+            # below even when an unusual interpreter has been found.
+            $data = $INTERPRETERS->value('php');
         } else {
             my $pinter = 0;
             if ($interpreter =~ m,^/,) {
                 # Check if the package ships the interpreter (and it is
                 # executable).
                 my $interfile = substr $interpreter, 1;
-                my $index_info = $info->index($interfile);
-                $pinter = 1 if $index_info && ($index_info->operm & 0111);
+                $pinter = 1 if $executable{$interfile};
+            } elsif (defined $calls_env) {
+                for my $dir (qw(usr/bin bin)) {
+                    if ($executable{"$dir/$interpreter"}) {
+                        $pinter = 1;
+                        last;
+                    }
+                }
             }
             script_tag('unusual-interpreter', $filename, "#!$interpreter")
               unless $pinter;
+
         }
 
         # Check for obsolete perl libraries
@@ -429,7 +444,9 @@ sub run {
                 $depends = $base;
             }
             if ($depends && !$all_parsed->implies($depends)) {
-                if ($base =~ /^(python|ruby|[mg]awk)$/) {
+                if ($base =~ /^php/) {
+                    tag 'php-script-but-no-php-cli-dep', $filename;
+                } elsif ($base =~ /^(python|ruby|[mg]awk)$/) {
                     tag("$base-script-but-no-$base-dep", $filename);
                 } elsif ($base eq 'csh' && $filename =~ m,^etc/csh/login\.d/,){
                     # Initialization files for csh.
@@ -456,9 +473,7 @@ sub run {
             unshift(@depends, $data->[1]) if length $data->[1];
             my $depends = join(' | ',  @depends);
             unless ($all_parsed->implies($depends)) {
-                if ($base eq 'php') {
-                    tag 'php-script-but-no-phpX-cli-dep', $filename;
-                } elsif ($base =~ /^(wish|tclsh)/) {
+                if ($base =~ /^(wish|tclsh)/) {
                     tag "$1-script-but-no-$1-dep", $filename;
                 } else {
                     tag 'missing-dep-for-interpreter', "$base => $depends",
@@ -470,9 +485,7 @@ sub run {
             my $depends = $data->[3];
             $depends =~ s/\$1/$version/g;
             unless ($all_parsed->implies($depends)) {
-                if ($base =~ /^php/) {
-                    tag 'php-script-but-no-phpX-cli-dep', $filename;
-                } elsif ($base =~ /^(python|ruby)/) {
+                if ($base =~ /^(python|ruby)/) {
                     tag "$1-script-but-no-$1-dep", $filename;
                 } else {
                     tag 'missing-dep-for-interpreter', "$base => $depends",
@@ -511,8 +524,7 @@ sub run {
     # normal scripts above, because there were just enough differences to
     # make a shared function awkward.
 
-    my %added_diversions;
-    my %removed_diversions;
+    my (%added_diversions, %removed_diversions, %dh_cmd_substs);
     my $expand_diversions = 0;
     while (<$ctrl_fd>) {
         chop;
@@ -619,6 +631,11 @@ sub run {
             if (/\#DEBHELPER\#/) {
                 tag 'maintainer-script-has-unexpanded-debhelper-token', $file;
             }
+            if (/^# Automatically added by (\S+)/) {
+                my $dh_cmd = $1;
+                tag 'debhelper-autoscript-in-maintainer-scripts', $dh_cmd
+                  if not $dh_cmd_substs{$dh_cmd}++;
+            }
 
             next if m,^\s*$,;  # skip empty lines
             next if m,^\s*\#,; # skip comment lines
@@ -1096,6 +1113,13 @@ sub run {
         }
     }
 
+    my @entries = $info->changelog ? $info->changelog->data : ();
+
+    if (@entries == 1) {
+        tag 'new-package-should-not-package-python2-module', $pkg
+          if $pkg =~ /^python-/;
+    }
+
     return;
 }
 
@@ -1205,7 +1229,7 @@ sub script_is_evil_and_wrong {
 sub check_script_syntax {
     my ($interpreter, $path) = @_;
     my $fs_path = $path->fs_path;
-    my $pid = fork();
+    my $pid = do_fork();
     if ($pid == 0) {
         open(STDOUT, '>', '/dev/null');
         open(STDERR, '>&', \*STDOUT);
