@@ -36,19 +36,6 @@ use Lintian::Util qw(fail strip);
 # one of the following names.
 my $HWCAP_DIRS = Lintian::Data->new('shared-libs/hwcap-dirs');
 
-# The following architectures should always have a STACK setting in shared
-# libraries to disable executable stack.  Other architectures don't always add
-# this section and therefore can't be checked.
-my %stack_arches = map { $_ => 1 }qw(
-  alpha
-  amd64
-  i386
-  m68k
-  powerpc
-  s390
-  sparc
-);
-
 # List of symbols file meta-fields.
 my %symbols_meta_fields = map { $_ => 1 }qw(
   Build-Depends-Package
@@ -77,7 +64,13 @@ sub run {
         next if not $file->is_file;
         my $fileinfo = $file->file_info;
         if ($fileinfo =~ m/^[^,]*\bELF\b/ && $fileinfo =~ m/shared object/) {
-            $sharedobject{$file} = 1;
+            my $perm = $file->operm;
+            my $debug = defined $objdump->{$file}{DEBUG};
+            if ($debug and $perm & 0111 and $file !~ m/\.so(?:\.|$)/) {
+                # position-independent executable
+            } else {
+                $sharedobject{$file} = 1;
+            }
         }
     }
 
@@ -151,20 +144,16 @@ sub run {
                 # special.
                 tag 'shlib-with-executable-bit', $cur_file, $perms
                   unless ($objdump->{$cur_file}{INTERP}
-                    or $cur_file =~ m,^lib(?:32|64)?/ld-[\d.]+\.so$,);
+                    or $cur_file =~ m,^lib.*/ld-[\d.]+\.so$,);
             } elsif ($perm != 0644) {
                 tag 'shlib-with-bad-permissions', $cur_file, $perms;
             }
 
-            # executable stack.  We can only warn about a missing
-            # section on some architectures.  Only warn if there's an
-            # Architecture field; if that's missing, we'll already be
-            # complaining elsewhere.
+            # executable stack.
             if (not defined $objdump->{$cur_file}{'PH'}{STACK}) {
                 if (defined $info->field('architecture')) {
                     my $arch = $info->field('architecture');
-                    tag 'shlib-without-PT_GNU_STACK-section', $cur_file
-                      if $stack_arches{$arch};
+                    tag 'shlib-without-PT_GNU_STACK-section', $cur_file;
                 }
             } elsif ($objdump->{$cur_file}{'PH'}{STACK}{flags} ne 'rw-'){
                 tag 'shlib-with-executable-stack', $cur_file;
